# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

import collections
import json
import threading
from typing import List, Any
import aiohttp
import urllib.parse


from .auth import Domains, Session
from .countries import get_country_code, get_region_by_country
from .consts import (
    APP_VERSION_CODE, APP_VERSION_NAME, PROJECT_TYPE,
    PROTOCOL_VERSION, REGION_URLS, ROBOT_PROPERTIES, TENANT_ID,
    Language, Region
)
from .device import Device, DeviceProperties
from .exception import KarcherHomeAccessDenied, KarcherHomeException, handle_error_code
from .map import Map
from .mqtt import MqttClient, get_device_topic_property_get_reply, get_device_topics
from .utils import (
    decrypt, decrypt_map, encrypt, get_nonce, get_random_string,
    get_timestamp, get_timestamp_ms, is_email, md5
)


class KarcherHome:
    """Main class to access Karcher Home Robots API"""

    @classmethod
    async def create(cls, country: str = 'GB', language: Language = Language.EN):
        """Create Karcher Home Robots API instance"""

        self = KarcherHome()
        self._country = country.upper()
        self._base_url = REGION_URLS[get_region_by_country(self._country)]
        self._language = language

        d = await self.get_urls()
        # Update base URLs
        if d.app_api != '':
            self._base_url = d.app_api
        if d.mqtt != '':
            self._mqtt_url = d.mqtt

        return self

    def __init__(self):
        """Initialize Karcher Home Robots API"""

        super().__init__()
        self._country = 'US'
        self._base_url = REGION_URLS[Region.US]
        self._mqtt_url = None
        self._language = Language.EN
        self._session = None
        self._mqtt = None
        self._device_props = {}
        self._wait_events = {}

    async def _request(self, method: str, url: str, **kwargs) -> aiohttp.ClientResponse:
        session = aiohttp.ClientSession()
        # TODO: Fix SSL
        # requests.packages.urllib3.disable_warnings()
        # session.skip = False

        headers = {}
        if kwargs.get('headers') is not None:
            headers = kwargs['headers']

        headers['User-Agent'] = 'Android_' + TENANT_ID
        auth = ''
        if self._session is not None and self._session.auth_token != '':
            auth = self._session.auth_token
            headers['authorization'] = auth
        if self._session is not None and self._session.user_id != '':
            headers['id'] = self._session.user_id
        headers['tenantId'] = TENANT_ID

        # Sign request
        nonce = get_nonce()
        ts = str(get_timestamp())
        data = ''
        if method == 'GET':
            params = kwargs.get('params') or {}
            if type(params) == str:
                params = urllib.parse.parse_qs(params)
            buf = urllib.parse.urlencode(params)
            data = buf
            kwargs['params'] = buf
        elif method == 'POST' or method == 'PUT':
            v = params = kwargs.get('json') or {}
            if type(v) == dict:
                v = collections.OrderedDict(v.items())
                for key, val in v.items():
                    data += key
                    if val is None:
                        data += 'null'
                    elif type(val) == str:
                        data += val
                    elif type(val) == dict:
                        data += json.dumps(val, separators=(',', ':'))
                    else:
                        data += str(val)
                kwargs['json'] = v

        headers['sign'] = md5(auth + ts + nonce + data)
        headers['ts'] = ts
        headers['nonce'] = nonce

        kwargs['headers'] = headers
        kwargs['verify_ssl'] = False
        return await session.request(method, self._base_url + url, **kwargs)

    async def _download(self, url) -> bytes:
        session = aiohttp.ClientSession()
        headers = {
            'User-Agent': 'Android_' + TENANT_ID,
        }

        resp = await session.get(url, headers=headers)
        if resp.status != 200:
            raise KarcherHomeException(-1,
                                       'HTTP error: ' + str(resp.status_code))

        return await resp.content.read(-1)

    async def _process_response(self, resp: aiohttp.ClientResponse, prop=None) -> Any:
        if resp.status != 200:
            raise KarcherHomeException(-1,
                                       'HTTP error: ' + str(resp.status))
        data = await resp.json()
        # Check for error response
        if data['code'] != 0:
            handle_error_code(data['code'], data['msg'])
        # Check for empty response
        if 'result' not in data:
            return None
        # Handle special response types
        result = data['result']
        if type(result) == str:
            raise KarcherHomeException(-2, 'Invalid response: ' + result)
        if prop is not None:
            return json.loads(decrypt(result[prop]))
        return result

    def _mqtt_connect(self, wait_for_connect=False):
        if self._session is None \
                or self._session.mqtt_token == '' or self._session.user_id == '':
            raise KarcherHomeAccessDenied('Not authorized')

        if self._mqtt is not None:
            return

        u = urllib.parse.urlparse("//" + self._mqtt_url)

        self._mqtt = MqttClient(
            host=u.hostname,
            port=u.port,
            username=self._session.user_id,
            password=self._session.mqtt_token)

        # Special logic for waiting for connection
        event = None
        if wait_for_connect:
            event = threading.Event()
            self._mqtt.on_connect = lambda: event.set()

        self._mqtt.connect()
        self._mqtt.on_message = self._process_mqtt_message

        if wait_for_connect:
            event.wait()
            self._mqtt.on_connect = None

    async def get_urls(self) -> Domains:
        """Get URLs for API and MQTT."""

        resp = await self._request('GET', '/network-service/domains/list', params={
            'tenantId': TENANT_ID,
            'productModeCode': PROJECT_TYPE,
            'version': PROTOCOL_VERSION,
        })

        d = await self._process_response(resp, 'domain')
        return Domains(**d)

    async def login(self, username, password, register_id=None) -> Session:
        """Login using provided credentials."""

        if register_id is None or register_id == '':
            register_id = get_random_string(19)

        if not is_email(username):
            username = '86-' + username

        resp = await self._request('POST', '/user-center/auth/login', json={
            'tenantId': TENANT_ID,
            'lang': str(self._language),
            'token': None,
            'userId': None,
            'password': encrypt(password),
            'username': encrypt(username),
            'authcode': None,
            'projectType': PROJECT_TYPE,
            'versionCode': APP_VERSION_CODE,
            'versionName': APP_VERSION_NAME,
            'phoneBrand': encrypt('xiaomi_mi 9'),
            'phoneSys': 1,
            'noticeSetting': {
                'andIpad': register_id,
                'android': register_id,
            },
        })

        d = await self._process_response(resp)
        self._session = Session(**d)
        self._session.register_id = register_id

        return self._session

    def login_token(
            self,
            auth_token: str,
            mqtt_token: str,
            register_id=None) -> Session:
        """Login using provided tokens."""

        if register_id is None or register_id == '':
            register_id = get_random_string(19)

        self._session = Session.from_token(auth_token, mqtt_token)
        self._session.register_id = register_id

        return self._session

    async def logout(self):
        """End current session.

        This will also reset the session object.
        """
        if self._session is None \
                or self._session.auth_token == '' or self._session.user_id == '':
            self._session = None
            return

        await self._process_response(await self._request(
            'POST', '/user-center/auth/logout'))
        self._session = None

        if self._mqtt is not None:
            self._mqtt.disconnect()
            self._mqtt = None

    async def get_devices(self) -> List[Device]:
        """Get all user devices."""

        if self._session is None \
                or self._session.auth_token == '' or self._session.user_id == '':
            raise KarcherHomeAccessDenied('Not authorized')

        resp = await self._request(
            'GET',
            '/smart-home-service/smartHome/user/getDeviceInfoByUserId/'
            + self._session.user_id)

        return [Device(**d) for d in await self._process_response(resp)]

    async def get_map_data(self, dev: Device, map: int = 1):
        # <tenantId>/<modeType>/<deviceSn>/01-01-2022/map/temp/0046690461_<deviceSn>_1
        mapDir = TENANT_ID + '/' + dev.product_mode_code + '/' +\
            dev.sn + '/01-01-2022/map/temp/0046690461_' + \
            dev.sn + '_' + str(map)

        resp = await self._request('POST',
                                   '/storage-management/storage/aws/getAccessUrl',
                                   json={
                                       'dir': mapDir,
                                       'countryCode': get_country_code(self._country),
                                       'serviceType': 2,
                                       'tenantId': TENANT_ID,
                                   })

        d = await self._process_response(resp)
        downloadUrl = d['url']
        if 'cdnDomain' in d and d['cdnDomain'] != '':
            downloadUrl = 'https://' + d['cdnDomain'] + '/' + d['dir']

        d = await self._download(downloadUrl)
        data = decrypt_map(dev.sn, dev.mac, dev.product_id, d)
        if map == 1 or map == 2:
            return Map.parse(data)
        else:
            return json.loads(data)

    def subscribe_device(self, dev: Device):
        """Subscribe to device real-time events."""

        if self._session is None \
                or self._session.mqtt_token == '' or self._session.user_id == '':
            raise KarcherHomeAccessDenied('Not authorized')

        self._mqtt_connect()
        self._device_props[dev.sn] = DeviceProperties()
        self._mqtt.subscribe(get_device_topics(dev.product_id, dev.sn))

    def unsubscribe_device(self, dev: Device):
        """Unsubscribe from device real-time events."""

        if self._session is None \
                or self._session.mqtt_token == '' or self._session.user_id == '':
            return

        if self._mqtt is None or dev.sn not in self._device_props:
            return

        self._mqtt.unsubscribe(get_device_topics(dev.product_id, dev.sn))
        del self._device_props[dev.sn]

    def _process_mqtt_message(self, topic, msg):
        sn = None
        for s in self._device_props.keys():
            if '/' + s + '/' in topic:
                sn = s
                break

        if sn is None:
            # Ignore messages for devices we have not subscribed to
            if topic in self._wait_events:
                self._wait_events[topic].set()
            return

        if 'thing/event/property/post' in topic \
                or 'thing/event/cur_path/post' in topic \
                or 'upgrade/post' in topic:
            if topic in self._wait_events:
                self._wait_events[topic].set()
            return

        if 'thing/service/property/get_reply' in topic:
            data = json.loads(msg)
            if data['code'] != 0:
                # TODO: handle error
                return
            self._update_device_properties(sn, data['data'])
            if topic in self._wait_events:
                self._wait_events[topic].set()
            return

        if topic in self._wait_events:
            self._wait_events[topic].set()

    def _wait_for_topic(self, topic: str, timeout: float = 5):
        if self._mqtt is None:
            return

        if topic in self._wait_events:
            return

        event = threading.Event()
        self._wait_events[topic] = event

        event.wait(timeout)
        del self._wait_events[topic]

    def _update_device_properties(self, sn: str, data: dict):
        if sn not in self._device_props:
            return

        self._device_props[sn].update(data)
        self._device_props[sn].last_update_time = get_timestamp()

    def request_device_update(self, dev: Device):
        """Request device update."""

        if self._session is None \
                or self._session.mqtt_token == '' or self._session.user_id == '':
            raise KarcherHomeAccessDenied('Not authorized')

        self._mqtt_connect()
        self._mqtt.publish(
            "/mqtt/" + dev.product_id + '/' + dev.sn + "/thing/service/property/get",
            json.dumps({
                "method": "prop.get",
                "msgId": str(get_timestamp_ms()),
                "tenantId": TENANT_ID,
                "version": "3.0",
                "params": {
                    "property": ROBOT_PROPERTIES,
                },
            }))

    def get_device_properties(self, dev: Device) -> DeviceProperties:
        """Get device properties if it has subscription."""

        if dev.sn in self._device_props:
            return self._device_props[dev.sn]

        if self._session is None \
                or self._session.mqtt_token == '' or self._session.user_id == '':
            raise KarcherHomeAccessDenied('Not authorized')

        self._mqtt_connect(wait_for_connect=True)
        subscr = dev.sn not in self._device_props
        if subscr:
            self.subscribe_device(dev)
        self.request_device_update(dev)
        self._wait_for_topic(
            get_device_topic_property_get_reply(dev.product_id, dev.sn))

        props = self._device_props[dev.sn]

        if subscr:
            self.unsubscribe_device(dev)

        return props



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

from typing import Final

from .consts import Region


COUNTRIES = {
    'AD': 'AND',
    'AE': 'UAE',
    'AF': 'AFG',
    'AG': 'ANT',
    'AL': 'ALB',
    'AM': 'ARM',
    'AR': 'ARG',
    'AT': 'AUT',
    'AZ': 'AZE',
    'BB': 'BAR',
    'BD': 'BAN',
    'BE': 'BEL',
    'BG': 'BUL',
    'BH': 'BRN',
    'BN': 'BRU',
    'BO': 'BOL',
    'BR': 'BRA',
    'BS': 'BAH',
    'BT': 'BHU',
    'BY': 'BLR',
    'BZ': 'BIZ',
    'CA': 'CAN',
    'CH': 'SUI',
    'CL': 'CHI',
    'CN': 'CHN',
    'CO': 'COL',
    'CR': 'CRC',
    'CU': 'CUB',
    'CY': 'CYP',
    'CZ': 'CZE',
    'DE': 'GER',
    'DI': 'SLO',
    'DK': 'DEN',
    'DM': 'DMA',
    'DO': 'DOM',
    'EC': 'ECU',
    'EE': 'EST',
    'ES': 'ES',
    'FI': 'FIN',
    'FR': 'FRA',
    'GB': 'GBR',
    'GD': 'GRN',
    'GE': 'GEO',
    'GR': 'GRE',
    'GT': 'GUA',
    'GY': 'GUY',
    'HK': 'HKG',
    'HN': 'HON',
    'HR': 'CRO',
    'HT': 'HAI',
    'HU': 'HUN',
    'ID': 'INA',
    'IE': 'IRL',
    'IL': 'ISR',
    'IN': 'IND',
    'IQ': 'IRQ',
    'IR': 'IRI',
    'IS': 'ISL',
    'IT': 'ITA',
    'JM': 'JAM',
    'JO': 'JOR',
    'JP': 'JPN',
    'KG': 'KGZ',
    'KH': 'CAM',
    'KP': 'PRK',
    'KR': 'KOR',
    'KW': 'KUW',
    'KZ': 'KAZ',
    'LA': 'LAO',
    'LB': 'LIB',
    'LC': 'LCA',
    'LI': 'LIE',
    'LK': 'SRI',
    'LT': 'LTU',
    'LU': 'LUX',
    'LV': 'LAT',
    'MC': 'MAS',
    'MD': 'MDA',
    'ME': 'MNE',
    'MK': 'MKD',
    'MM': 'MYA',
    'MN': 'MGL',
    'MO': 'Macao',
    'MT': 'MLT',
    'MV': 'MDV',
    'MX': 'MEX',
    'NI': 'NCA',
    'NL': 'NED',
    'NO': 'NOR',
    'NP': 'NEP',
    'OM': 'OMA',
    'PA': 'PAN',
    'PE': 'PER',
    'PH': 'PHI',
    'PK': 'PAK',
    'PL': 'POL',
    'PT': 'POR',
    'PY': 'PAR',
    'QA': 'QAT',
    'RO': 'ROU',
    'RS': 'SRB',
    'RU': 'RUS',
    'SA': 'KSA',
    'SE': 'SWE',
    'SG': 'SIN',
    'SK': 'SVK',
    'SM': 'SMR',
    'SR': 'SUR',
    'SV': 'ESA',
    'SY': 'SYR',
    'TH': 'THA',
    'TJ': 'TJK',
    'TL': 'TLS',
    'TM': 'TKM',
    'TR': 'TUR',
    'TT': 'TRI',
    'UA': 'UKR',
    'US': 'USA',
    'UY': 'URU',
    'UZ': 'UZB',
    'VA': 'VCS',
    'VC': 'VIN',
    'VE': 'VEN',
    'VN': 'VIE',
    'YE': 'YEM',
    'BIH': 'BIH',
    'PLE': 'PLE',
    'SKN': 'SKN',
    'TWN': 'TWN',
}

COUNTRY_REGIONS: Final = {
    'AND': Region.EU,
    'UAE': Region.EU,
    'AFG': Region.CN,
    'ANT': Region.US,
    'ALB': Region.EU,
    'ARM': Region.EU,
    'ARG': Region.US,
    'AUT': Region.EU,
    'AZE': Region.EU,
    'BAR': Region.US,
    'BAN': Region.CN,
    'BEL': Region.EU,
    'BUL': Region.EU,
    'BRN': Region.EU,
    'BRU': Region.CN,
    'BOL': Region.US,
    'BRA': Region.US,
    'BAH': Region.US,
    'BHU': Region.CN,
    'BLR': Region.EU,
    'BIZ': Region.US,
    'CAN': Region.US,
    'SUI': Region.EU,
    'CHI': Region.US,
    'CHN': Region.CN,
    'COL': Region.US,
    'CRI': Region.US,
    'CUB': Region.US,
    'CYP': Region.EU,
    'CZE': Region.EU,
    'GER': Region.EU,
    'DEN': Region.EU,
    'DMA': Region.US,
    'DOM': Region.US,
    'ECU': Region.US,
    'EST': Region.EU,
    'ESP': Region.EU,
    'FIN': Region.EU,
    'FRA': Region.EU,
    'GBR': Region.EU,
    'GRN': Region.US,
    'GEO': Region.EU,
    'GRE': Region.EU,
    'GUA': Region.US,
    'GUY': Region.US,
    'HKG': Region.CN,
    'HON': Region.US,
    'CRO': Region.EU,
    'HAI': Region.US,
    'HUN': Region.EU,
    'INA': Region.CN,
    'IRL': Region.EU,
    'ISR': Region.EU,
    'IND': Region.CN,
    'IRQ': Region.EU,
    'IRI': Region.EU,
    'ISL': Region.EU,
    'ITA': Region.EU,
    'JAM': Region.US,
    'JOR': Region.EU,
    'JPN': Region.CN,
    'KEN': Region.EU,
    'KGZ': Region.CN,
    'CAM': Region.CN,
    'KOR': Region.CN,
    'KUW': Region.EU,
    'KAZ': Region.CN,
    'LAO': Region.CN,
    'LIB': Region.EU,
    'LCA': Region.US,
    'LIE': Region.EU,
    'SRI': Region.CN,
    'LTU': Region.EU,
    'LUX': Region.EU,
    'LAT': Region.EU,
    'MAS': Region.CN,
    'MON': Region.EU,
    'MDA': Region.EU,
    'MGL': Region.CN,
    'Macao': Region.CN,
    'MAC': Region.CN,
    'MEX': Region.US,
    'NED': Region.EU,
    'NOR': Region.EU,
    'NEP': Region.CN,
    'NZL': Region.US,
    'OMA': Region.EU,
    'PAN': Region.US,
    'PER': Region.US,
    'PHI': Region.CN,
    'PAK': Region.CN,
    'POL': Region.EU,
    'PUR': Region.US,
    'POR': Region.EU,
    'PAR': Region.US,
    'QAT': Region.EU,
    'ROU': Region.EU,
    'SRB': Region.EU,
    'RUS': Region.EU,
    'KSA': Region.EU,
    'SIN': Region.CN,
    'SVK': Region.EU,
    'SMR': Region.EU,
    'SUR': Region.US,
    'ESA': Region.US,
    'SYR': Region.EU,
    'THA': Region.CN,
    'TJK': Region.CN,
    'TLS': Region.CN,
    'TKM': Region.CN,
    'TUR': Region.EU,
    'TRI': Region.US,
    'UKR': Region.EU,
    'USA': Region.US,
    'URU': Region.US,
    'UZB': Region.CN,
    'VCS': Region.US,
    'VIN': Region.US,
    'VEN': Region.US,
    'VIE': Region.CN,
    'YEM': Region.EU,
    'BIH': Region.EU,
    'PLE': Region.EU,
    'SKN': Region.US,
    'TWN': Region.CN,
}


def get_country_code(country: str) -> str:
    """Get country code."""

    country = country.upper()
    if country in COUNTRIES:
        return COUNTRIES[country]
    return country


def get_region_by_country(country: str) -> Region:
    """Get region by country."""

    country = get_country_code(country)

    # Return region (US, EU or CN) based on country code
    if country in COUNTRY_REGIONS:
        return COUNTRY_REGIONS[country]
    return Region.EU



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

import base64
from dataclasses import dataclass, fields
import json


@dataclass(init=False)
class Domains:
    """Domains URLs class.

    This class represents a Karcher Home access URLs.
    """
    app_api: str
    mqtt: str

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            k = k.lower()
            if k in names:
                if k == 'app_api':
                    v = 'https://' + v
                setattr(self, k, v)


@dataclass(init=False)
class Session:
    """Authorized user session class.

    This class represents a Karcher Home authorized user session.
    """

    register_id: str
    user_id: str
    auth_token: str
    mqtt_token: str

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        if 'id' in kwargs:
            setattr(self, 'user_id', kwargs['id'])
        if 'data' in kwargs:
            kwargs = kwargs['data']
        for k, v in kwargs.items():
            k = k.lower()
            if k == 'auth':
                k = 'auth_token'
            elif k == 'emq_token':
                k = 'mqtt_token'
            if k in names:
                setattr(self, k, v)

    def reset(self):
        """Reset session."""
        self.user_id = ''
        self.auth_token = ''
        self.mqtt_token = ''

    @staticmethod
    def from_token(auth_token: str, mqtt_token: str):
        """Create session from auth and MQTT tokens."""
        sess = Session()

        # Get user ID from auth token
        token_data = auth_token.split(".")[1]
        missing_padding = len(token_data) % 4
        if missing_padding:
            token_data += '=' * (4 - missing_padding)
        data = json.loads(base64.b64decode(token_data))
        auth = json.loads(data["value"])

        sess.user_id = auth["id"]
        sess.auth_token = auth_token
        sess.mqtt_token = mqtt_token

        return sess



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

import base64
import hashlib
import random
import re
import string
import time
from typing import Final
import zlib
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .consts import TENANT_ID, Product


EMAIL_REGEX: Final = "^\\w+([-+.]\\w+)*@\\w+([-.]\\w+)*\\.\\w+([-.]\\w+)*$"


def get_random_string(length: int) -> str:
    letters = string.digits + string.ascii_lowercase + string.ascii_uppercase
    return ''.join(random.choice(letters) for i in range(length))


def get_random_device_id() -> str:
    return ''.join(random.choice('0123456789abcdef') for i in range(16))


def get_nonce() -> str:
    return get_random_string(32)


def get_timestamp() -> int:
    return int(time.time())


def get_timestamp_ms() -> int:
    return int(time.time() * 1000)


def get_enc_key() -> bytes:
    m = hashlib.md5()
    m.update(bytes(TENANT_ID, 'utf-8'))
    h = m.hexdigest()
    return bytes(h[8:24], 'utf-8')


def decrypt(data) -> str:
    cipher = Cipher(algorithms.AES128(get_enc_key()), modes.ECB())
    buf = cipher.decryptor().update(base64.b64decode(data))
    return str(buf[:-ord(buf[-1:])], 'utf-8')


def encrypt(data) -> str:
    cipher = Cipher(algorithms.AES128(get_enc_key()), modes.ECB())
    buf = bytes(data, 'utf-8')
    pad_len = 16 - (len(buf) % 16)
    buf = buf + bytes([pad_len]) * pad_len
    return base64.b64encode(cipher.encryptor().update(buf)).decode()


def get_map_enc_key(sn: str, mac: str, product_id: Product) -> bytes:
    sub_key = mac.replace(':', '').lower() + str(product_id.value)
    cipher = Cipher(algorithms.AES128(bytes(sub_key[0:16], 'utf-8')), modes.ECB())
    buf = bytes(sn + '+' + str(product_id.value) + '+' + sn, 'utf-8')
    pad_len = 16 - (len(buf) % 16)
    buf = buf + bytes([pad_len]) * pad_len

    key = base64.b64encode(cipher.encryptor().update(buf)).decode()

    m = hashlib.md5()
    m.update(bytes(key, 'utf-8'))
    h = m.hexdigest()
    return bytes(h[8:24], 'utf-8')


def decrypt_map(sn: str, mac: str, product_id: Product, data: bytes) -> bytes:
    key = get_map_enc_key(sn, mac, product_id)
    cipher = Cipher(algorithms.AES128(key), modes.ECB())
    buf = cipher.decryptor().update(base64.b64decode(data))
    try:
        return zlib.decompress(bytes.fromhex(str(buf[:-ord(buf[-1:])], 'utf-8')))
    except Exception:
        return bytes.fromhex(str(buf[:-ord(buf[-1:])], 'utf-8'))


def md5(data: str) -> str:
    m = hashlib.md5()
    m.update(bytes(data, 'utf-8'))
    return m.hexdigest()


def is_email(email: str) -> bool:
    return re.search(EMAIL_REGEX, email) is not None


def snake_case(value: str) -> str:
    first_underscore = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', value)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', first_underscore).lower()


def snake_case_fields(data: str) -> str:
    if type(data) is dict:
        n = {}
        for k, v in data.items():
            n[snake_case(k)] = snake_case_fields(v)
        return n
    elif type(data) is list:
        n = []
        for v in data:
            n.append(snake_case_fields(v))
        return n
    else:
        return data



from typing import List
import ssl
from paho.mqtt.client import Client, MQTTv311

from .utils import get_random_device_id


class MqttClient:
    def __init__(self, host, port, username, password):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._topics = []
        self._client = Client(
            client_id=self._username + '_' + get_random_device_id(),
            clean_session=True,
            protocol=MQTTv311)
        self.on_message = None
        self.on_connect = None

    def connect(self):
        # TODO validate certificate
        self._client.tls_set(cert_reqs=ssl.CERT_NONE)
        self._client.tls_insecure_set(True)
        self._client.username_pw_set(self._username, self._password)
        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        # TODO: add option to enable logging
        # self._client.on_log = lambda client, userdata, level, buf: print(buf)
        self._client.connect(self._host, self._port, 60)
        self._client.loop_start()

    def disconnect(self):
        self._client.loop_stop()
        self._client.disconnect()

    def _subscribe(self, topics):
        if len(topics) == 0:
            return
        t = []
        for topic in topics:
            t.append((topic, 0))
        self._client.subscribe(t)

    def subscribe(self, topics):
        t = []
        for topic in topics:
            if topic not in self._topics:
                self._topics.append(topic)
                t.append(topic)
        if self._client.is_connected():
            self._subscribe(t)

    def unsubscribe(self, topics):
        t = []
        for topic in topics:
            if topic in self._topics:
                self._topics.remove(topic)
                t.append(topic)
        if self._client.is_connected():
            self._client.unsubscribe(t)

    def publish(self, topic, payload):
        self._client.publish(topic, payload)

    def _on_connect(self, client, userdata, flags, rc):
        self._subscribe(self._topics)
        if self.on_connect is not None:
            self.on_connect()

    def _on_message(self, client, userdata, msg):
        if self.on_message is not None:
            self.on_message(msg.topic, msg.payload)

    def __del__(self):
        self.disconnect()


def get_device_topics(product_id: str, sn: str) -> List[str]:
    return [
        '/mqtt/' + product_id + '/' + sn + '/thing/event/property/post',
        '/mqtt/' + product_id + '/' + sn + '/thing/service/property/set_reply',
        get_device_topic_property_get_reply(product_id, sn),
        '/mqtt/' + product_id + '/' + sn + '/thing/service_invoke',
        '/mqtt/' + product_id + '/' + sn + '/thing/service_invoke_reply/#',
        '/mqtt/' + product_id + '/' + sn + '/thing/event/cur_path/post',
        '/mqtt/' + product_id + '/' + sn + '/ota/service/upgrade/set_reply',
        '/mqtt/' + product_id + '/' + sn + '/ota/service/upgrade/post',
        '/mqtt/' + product_id + '/' + sn + '/ota/service/upgrade/get_reply',
        '/mqtt/' + product_id + '/' + sn + '/ota/service/version/post',
    ]


def get_device_topic_property_get_reply(product_id: str, sn: str) -> str:
    return '/mqtt/' + product_id + '/' + sn + '/thing/service/property/get_reply'



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

class KarcherHomeException(Exception):
    """Exception raised for errors in the Karcher Home Robots library.

    Attributes:
        code -- error code
        message -- explanation of the error
    """

    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(self.message)


class KarcherHomeAccessDenied(KarcherHomeException):
    """Exception raised when user has no access for resource.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(608, self.message)


class KarcherHomeInvalidAuth(KarcherHomeException):
    """Exception raised when wrong credentials are provided."""

    def __init__(self):
        super().__init__(620, 'The username or password is incorrect')


class KarcherHomeTokenExpired(KarcherHomeException):
    """Exception raised when token has been expired."""

    def __init__(self):
        super().__init__(609,
                         'Unauthorized or authorization expired, please log in again')


def handle_error_code(code, message):
    if code == 608:
        raise KarcherHomeAccessDenied('Forbidden')
    elif code == 609:
        raise KarcherHomeTokenExpired()
    elif code == 613:
        raise KarcherHomeException(613, 'Invalid token')
    elif code == 620:
        raise KarcherHomeInvalidAuth()
    else:
        raise KarcherHomeException(code, message)



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

"""Compat layer for homeassistant."""
from enum import Enum, auto


class VacuumState(Enum):
    """Vacuum state enum.

    This offers a simplified API to the vacuum state.

    # TODO: the interpretation of simplified state should be done downstream.
    """

    Unknown = auto()
    Cleaning = auto()
    Returning = auto()
    Idle = auto()
    Docked = auto()
    Paused = auto()
    Error = auto()



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

from dataclasses import dataclass, fields
from enum import Enum
import json
from typing import Any, List

from .consts import Product
from .utils import snake_case


class DeviceStatus(int, Enum):
    """Device status enum."""
    Offline = 0
    Online = 1


@dataclass(init=False)
class DeviceVersion:
    """Device version class.

    This class represents a Karcher Home device version.
    """
    package_type: str = ''
    version: int = 0
    version_name: str = ''
    ctrl_version: str = ''

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            k = snake_case(k)
            if k in names:
                setattr(self, k, v)


@dataclass
class DevicePropertiesPrivacy:
    """Device privacy properties class."""

    ai_recognize: int = 0
    dirt_recognize: int = 0
    pet_recognize: int = 0
    carpet_turbo: int = 0
    carpet_avoid: int = 0
    carpet_show: int = 0
    auto_upgrade: int = 0
    map_uploads: int = 0
    record_uploads: int = 0


@dataclass
class DevicePropertiesNetwork:
    """Device network status properties class."""

    rssi: str = ''
    loss: int = 0
    ping: int = 0
    ip: str = ''
    mac: str = ''


@dataclass
class DevicePropertiesOrderTotal:
    """Device order total properties class."""

    enable: int = 0
    total: int = 0


@dataclass
class DevicePropertiesQuiet:
    """Device quite status properties class."""

    begin_time: int = 0
    end_time: int = 0
    quiting: int = 0


@dataclass(init=False)
class DeviceProperties:
    """Device properties class.

    This class represents a Karcher Home detailed device properties.
    """

    firmware: str = ''
    firmware_code: int = 0
    status: int = 0
    fault: int = 0
    wind: int = 0
    water: int = 0
    mode: int = 0
    quantity: int = 0
    alarm: int = 0
    volume: int = 0
    hypa: int = 0
    main_brush: int = 0
    side_brush: int = 0
    mop_life: int = 0
    net_status: DevicePropertiesNetwork = DevicePropertiesNetwork()
    repeat_state: int = 0
    tank_state: int = 0
    cloth_state: int = 0
    sweep_type: int = 0
    mop_route: int = 0
    time_zone: int = 0
    language: int = 0
    cleaning_time: int = 0
    cleaning_area: int = 0
    custom_type: int = 0
    sound: int = 0
    work_mode: int = 0
    tank_shake: int = 0
    shake_shift: int = 0
    electrolysis: int = 0
    station_act: int = 0
    charge_state: int = 0
    back_to_wash: int = 0
    break_charging: int = 0
    order_total: DevicePropertiesOrderTotal = DevicePropertiesOrderTotal()
    memory_map: int = 0
    current_map_id: int = 0
    map_num: int = 0
    has_new_map: int = 0
    build_map: int = 0
    quiet_is_open: int = 0
    quiet_begin_time: int = 0
    quiet_end_time: int = 0
    broken_clean: int = 0
    privacy: DevicePropertiesPrivacy = DevicePropertiesPrivacy()
    cur_path: List[float]
    dust_action: int = 0
    voice_type: int = 0
    quiet_status: DevicePropertiesQuiet = DevicePropertiesQuiet()
    last_update_time: int = 0

    def __init__(self, **kwargs):
        setattr(self, 'cur_path', [])
        self.update(kwargs)

    def update(self, data: dict[str, Any]) -> bool:
        """Update device properties."""

        updated = False
        names = set([f.name for f in fields(self)])
        for k, v in data.items():
            if k in names:
                if k == 'firmware_code':
                    v = int(v)
                if v != getattr(self, k):
                    setattr(self, k, v)
                    updated = True
        return updated


@dataclass(init=False)
class Device:
    """Device class.

    This class represents a Karcher Home device.
    """

    device_id: str
    sn: str
    mac: str
    nickname: str
    versions: List[DeviceVersion]
    status: DeviceStatus
    is_default: bool
    is_selected: bool
    is_shared: bool
    online_time: int
    photo_url: str
    product_id: Product
    product_mode_code: str
    bind_time: int
    room_id: str

    def __init__(self, **kwargs):
        # Set default values
        for k, v in [('isSelected', False), ('isShare', False), ('isDefault', False)]:
            if k not in kwargs:
                kwargs[k] = v

        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            k = snake_case(k)
            if k == 'is_share':
                k = 'is_shared'
            if k in names:
                if k == 'status':
                    v = DeviceStatus(v)
                if k == 'product_id':
                    v = Product(v)
                elif k == 'versions':
                    v = json.loads(v)
                    v = [DeviceVersion(**x) for x in v]
                setattr(self, k, v)

    def is_online(self):
        """Get device status."""

        return self.status == DeviceStatus.Online



# -----------------------------------------------------------
# Copyright (c) 2023 Lauris BH
# SPDX-License-Identifier: MIT
# -----------------------------------------------------------

from enum import Enum


class Region(str, Enum):
    """Region enum.

    Cloud region.
    """
    EU = 'eu'
    US = 'us'
    CN = 'cn'


class Language(int, Enum):
    """Language enum.

    Language of the API.
    """
    CN = 1
    EN = 2
    ES = 3
    DE = 4
    FR = 5
    PL = 6
    IT = 7
    TR = 12
    CS = 15
    NL = 16
    SV = 17

    def __str__(self):
        if self == Language.CN:
            return 'cn'
        elif self == Language.EN:
            return 'en'
        elif self == Language.ES:
            return 'es'
        elif self == Language.DE:
            return 'de'
        elif self == Language.FR:
            return 'fr'
        elif self == Language.PL:
            return 'pl'
        elif self == Language.IT:
            return 'it'
        elif self == Language.TR:
            return 'tr'
        elif self == Language.CS:
            return 'cs'
        elif self == Language.NL:
            return 'nl'
        elif self == Language.SV:
            return 'sv'
        else:
            return 'en'


class Product(str, Enum):
    """Product model enum."""

    RCV3 = '1528986273083777024'
    RCV5 = '1540149850806333440'
    RCF5 = '1599715149861306368'


REGION_URLS = {
    Region.EU: 'https://eu-appaiot.3irobotix.net',
    Region.US: 'https://us-appaiot.3irobotix.net',
    Region.CN: 'https://cn-appaiot.3irobotix.net',
}

REGION_INDEX = {
    Region.EU: 1,
    Region.US: 2,
    Region.CN: 0,
}

ROBOT_PROPERTIES = [
    'status',
    'firmware_code',
    'firmware',
    'fault',
    'mode',
    'wind',
    'water',
    'repeat_state',
    'charge_state',
    'quantity',
    'work_mode',
    'sweep_type',
    'build_map',
    'cleaning_area',
    'cleaning_time',
    'current_map_id',
    'custom_type',
    'privacy',
    'alarm',
    'volume',
    'tank_state',
    'cloth_state',
    'mop_route',
    'map_num',
    'language',
    'voice_type',
    'quiet_status',
    'quiet_is_open'
]

TENANT_ID = '1528983614213726208'
PROJECT_TYPE = 'android_iot.karcher'
PROTOCOL_VERSION = 'v1'
APP_VERSION_CODE = 10004
APP_VERSION_NAME = '1.0.4'
